use oauth2::{
    basic::BasicClient, AuthUrl, AuthorizationCode, ClientId, ClientSecret, CsrfToken,
    PkceCodeChallenge, PkceCodeVerifier, RedirectUrl, RevocationUrl, Scope, TokenUrl,
};
use reqwest::Url;

use crate::{errors::ApiError, settings::Settings};

use super::OauthProvider;

#[derive(Debug, Clone)]
pub struct DiscordOauthProvider {
    client: BasicClient,
}

impl DiscordOauthProvider {
    pub fn new(settings: &Settings) -> Self {
        let redirect_url = format!("{}/v1/api/auth/callback/discord", settings.web_address());
        let auth_url = AuthUrl::new("https://discord.com/oauth2/authorize".to_string())
            .expect("Invalid authorization endpoint URL");
        let token_url = TokenUrl::new("https://discord.com/api/oauth2/token".to_string())
            .expect("Invalid token endpoint URL");
        let revocation_url =
            RevocationUrl::new("https://discord.com/api/oauth2/token/revoke".to_string())
                .expect("Invalid revocation URL");
        let client = BasicClient::new(
            ClientId::new(settings.discord.client_id.clone()),
            Some(ClientSecret::new(settings.discord.client_secret.clone())),
            auth_url,
            Some(token_url),
        )
        .set_redirect_uri(RedirectUrl::new(redirect_url).expect("Invalid redirect URL"))
        .set_revocation_uri(revocation_url);
        Self { client }
    }
}

impl OauthProvider for DiscordOauthProvider {
    fn auth_and_csrf(&self) -> (Url, CsrfToken, PkceCodeVerifier) {
        let (pkce_challenge, pkce_verifier) = PkceCodeChallenge::new_random_sha256();
        let (auth_url, csrf_token) = self
            .client
            .authorize_url(CsrfToken::new_random)
            .add_scopes(["identify", "openid", "email"].map(|v| Scope::new(v.to_string())))
            .set_pkce_challenge(pkce_challenge)
            .url();
        (auth_url, csrf_token, pkce_verifier)
    }

    async fn token(
        &self,
        code: String,
        verifier: PkceCodeVerifier,
    ) -> Result<super::Token, ApiError> {
        self.client
            .exchange_code(AuthorizationCode::new(code))
            .set_pkce_verifier(verifier)
            .request_async(oauth2::reqwest::async_http_client)
            .await
            .map_err(|e| ApiError::TokenError(format!("{e:?}")))
    }
}
